import logging
import logging.handlers


class Logger(object):

	__instance = None
	__initialized = False
	__format = '%(asctime)s :: %(levelname)s: %(message)s'

	@staticmethod
	def init(level: str='info'):
		if Logger.__instance is None:
			Logger.initialize(level)
			Logger.__initialized = True
		return Logger.__instance

	@staticmethod
	def initialize(level: str):
		level = Logger.translate_log_level(level)

		Logger.__instance = logging.getLogger('smashggpy')
		Logger.__instance.setLevel(level)

		console_handler = logging.StreamHandler()
		console_handler.setLevel(level)

		formatter = logging.Formatter(Logger.__format)
		console_handler.setFormatter(formatter)

		Logger.__instance.addHandler(console_handler)
		Logger.__instance.info('initialized Logger')

	@staticmethod
	def set_log_level(level: str='info'):
		Logger.initialize(level)

	@staticmethod
	def translate_log_level(level: str='info'):
		if level not in ('info', 'warning', 'debug', 'error', 'critical'):
			raise Exception('invalid value for log level: {}'.format(level))

		if level is 'info': return logging.INFO
		elif level is 'error': return logging.ERROR
		elif level is 'warning': return logging.WARNING
		elif level is 'critical': return logging.CRITICAL
		elif level is 'debug': return logging.DEBUG

	@staticmethod
	def verify_initialized():
		if Logger.__initialized is not True:
			raise Exception('Logger class is not initialized!')

	@staticmethod
	def get_instance():
		Logger.verify_initialized()
		return Logger.__instance

	@staticmethod
	def add_handler(handler):
		Logger.verify_initialized()
		Logger.get_instance().addHandler(handler)
	
	@staticmethod
	def info(message):
		Logger.__instance.info(message)

	@staticmethod
	def debug(message):
		Logger.__instance.debug(message)

	@staticmethod
	def critical(message):
		Logger.__instance.critical(message)

	@staticmethod
	def error(message):
		Logger.__instance.error(message)

	@staticmethod
	def warning(message):
		Logger.__instance.warning(message)



class Organizer(object):

    def __init__(self, id, email, phone, twitter, info=None):
        self.id = id
        self.email = email
        self.phone = phone
        self.twitter = twitter
        self.info = info

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((self.id, self.email, self.phone, self.twitter, self.info))

    @staticmethod
    def parse(data):
        assert (data is not None), 'Organizer.parse must not have a None data parameter'
        assert ('ownerId' in data), 'Organizer.parse must have an ownerId property in data parameter'
        assert ('contactEmail' in data), 'Organizer.parse must have a contactEmail property in data parameter'
        assert ('contactPhone' in data), 'Organizer.parse must have a contactInfo property in data parameter'
        assert ('contactTwitter' in data), 'Organizer.parse must have a contactTwitter property in data parameter'
        # assert ('contactInfo' in data), 'Organizer.parse must have a contactInfo property in data parameter
        return Organizer(
            data['ownerId'],
            data['contactEmail'],
            data['contactPhone'],
            data['contactTwitter'],
            # data['contactInfo']
        )

    # GETTERS
    def get_id(self):
        return self.id

    def get_email(self):
        return self.email

    def get_phone(self):
        return self.phone

    def get_twitter(self):
        return self.twitter

    def get_info(self):
        return self.info




class Player(object):

    def __init__(self, tag, entrant_id, attendee_ids):
        self.tag = tag
        self.entrant_id = entrant_id
        self.attendee_ids = attendee_ids

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    @staticmethod
    def parse(tag, data):
        

        entrant_id = data['entrant']['id'] if data['entrant'] is not None else None
        participant_data = data['entrant']['participants']
        attendee_ids = [participant['id'] for participant in participant_data]
        return Player(
            tag,
            entrant_id,
            attendee_ids
        )

    # GETTERS
    def get_tag(self):
        return self.tag

    def get_entrant_id(self):
        return self.entrant_id

    def get_attendee_ids(self):
        return self.attendee_ids



import smashggpy.queries.Schema as schema

get_tournament = """
query TournamentQuery($id: ID!){{
    tournament(id: $id){{
        {0}
    }}
}}
""".format(schema.tournament_schema)

get_tournament_by_slug = """
query TournamentQuery($slug: String){{
    tournament(slug: $slug){{
        {0}
    }}
}}
""".format(schema.tournament_schema)


get_tournament_events = """
query TournamentEvents($id: ID!){{
    tournament(id: $id){{
        events{{
            {0}
        }}
    }}
}}
""".format(schema.event_schema)

get_tournament_phases = """
query TournamentPhases($id: ID!){{
    tournament(id: $id){{
        events{{
            phases{{
                {0}
            }}
        }}
    }}
}}
""".format(schema.phase_schema)

get_tournament_phase_groups = """
query TournamentPhaseGroups($id: ID!){{
    tournament(id: $id){{
        events{{
            phaseGroups{{
                {0}
            }}
        }}
    }}
}}
""".format(schema.phase_group_schema)



import smashggpy.queries.Schema as schema

phase_by_id = """
query PhaseQuery($id: ID!){{
    phase(id: $id){{
        {0}
    }}
}}
""".format(schema.phase_schema)

phase_phase_groups = """
query PhasePhaseGroups($id: ID!, $page: Int, $perPage: Int, $sortBy: String, $entrantIds: [ID], $filter: PhaseGroupPageQueryFilter){{
   phase(id: $id){{
        phaseGroups(query: {{
            page: $page,
            perPage: $perPage,
            sortBy: $sortBy,
            entrantIds: $entrantIds,
            filter: $filter
        }})
        {{
            pageInfo{{
                totalPages
            }}
            nodes{{
                {0}
            }}  
        }}
    }}
}}
""".format(schema.phase_group_schema)


import re

DISPLAY_SCORE_REGEX = re.compile('^([\S\s]*) ([0-9]{1,3}) - ([\S\s]*) ([0-9]{1,3})$')

class GGSet(object):

    def __init__(self, id, event_id, phase_group_id, display_score, full_round_text, round,
                 started_at, completed_at, winner_id, total_games, state,
                 player1, player2, score1, score2):
        self.id = id
        self.event_id = event_id
        self.phase_group_id = phase_group_id
        self.display_score = display_score
        self.full_round_text = full_round_text
        self.round = round
        self.started_at = started_at
        self.completed_at = completed_at
        self.winner_id = winner_id
        self.total_games = total_games
        self.state = state
        self.player1 = player1
        self.player2 = player2
        self.score1 = score1
        self.score2 = score2

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.event_id, self.phase_group_id, self.display_score, self.full_round_text,
                     self.round, self.started_at, self.completed_at, self.winner_id, self.total_games,
                     self.state, self.player1, self.player1, self.score1, self.score2))

    @staticmethod
    def parse(data):
        assert (data is not None), 'GGSet.parse cannot have a none data parameter'
        assert ('id' in data), 'GGSet.parse must have an id property in data parameter'
        assert ('eventId' in data), 'GGset.parse must have a eventId property in data parameter'
        assert ('phaseGroupId' in data), 'GGset.parse must have a phaseGroupId property in data parameter'
        assert ('displayScore' in data), 'GGset.parse must have a displayScore property in data parameter'
        assert ('fullRoundText' in data), 'GGset.parse must have a fullRoundText property in data parameter'
        assert ('round' in data), 'GGset.parse must have a round property in data parameter'
        assert ('startedAt' in data), 'GGset.parse must have a startedAt property in data parameter'
        assert ('completedAt' in data), 'GGset.parse must have a completedAt property in data parameter'
        assert ('winnerId' in data), 'GGset.parse must have a winnerId property in data parameter'
        assert ('totalGames' in data), 'GGset.parse must have a totalGames property in data parameter'
        assert ('state' in data), 'GGset.parse must have a state property in data parameter'

        display_score_parsed = GGSet.parse_display_score(data['displayScore'])
        assert ('p1_tag' in data), 'GGSet.parse must resolve p1_tag property in display score'
        assert ('p2_tag' in data), 'GGSet.parse must resolve p2_tag property in display score'
        assert ('p1_score' in data), 'GGSet.parse must resolve p1_score property in display score'
        assert ('p2_score' in data), 'GGSet.parse must resolve p2_score property in display score'

        return GGSet(
            data['id'],
            data['eventId'],
            data['phaseGroupId'],
            data['displayScore'],
            data['fullRoundText'],
            data['round'],
            data['startedAt'],
            data['completedAt'],
            data['winnerId'],
            data['totalGames'],
            data['state'],
            display_score_parsed['p1_tag'],
            display_score_parsed['p2_tag'],
            display_score_parsed['p1_score'],
            display_score_parsed['p2_score']
        )

    @staticmethod
    def parse_display_score(display_score: str):
        ret = {
            'p1_tag': None,
            'p1_score': -1,
            'p2_tag': None,
            'p2_score': -1
        }

        if display_score is not None or display_score != '':
            matches = DISPLAY_SCORE_REGEX.match(display_score)
            if matches is not None:
                ret['p1_tag'] = matches.group(1)
                ret['p1_score'] = matches.group(2)
                ret['p2_tag'] = matches.group(3)
                ret['p2_score'] = matches.group(4)

        return ret
    
    def get_is_completed(self):
        return self.completed_at is not None

    # GETTERS
    def get_id(self):
        return self.id

    def get_event_id(self):
        return self.event_id

    def get_phase_group_id(self):
        return self.phase_group_id

    def get_display_score(self):
        return self.display_score

    def get_full_round_text(self):
        return self.full_round_text

    def get_round(self):
        return self.round

    def get_started_at(self):
        return self.started_at

    def get_completed_at(self):
        return self.completed_at

    def get_winner_id(self):
        return self.winner_id

    def get_total_games(self):
        return self.total_games

    def get_state(self):
        return self.state

    def get_player1(self):
        return self.player1

    def get_player2(self):
        return self.player2

    def get_score1(self):
        return self.score1

    def get_score2(self):
        return self.score2


from smashggpy.models.Player import Player


import smashggpy.queries.Event_Queries as queries
from smashggpy.util.Logger import Logger
from smashggpy.common.Common import flatten, validate_data
from smashggpy.util.NetworkInterface import NetworkInterface as NI
from smashggpy.util.ThreadFactory import ThreadFactory
from smashggpy.common.Exceptions import \
    DataPullException, NoEventDataException, NoPhaseGroupDataException, NoPhaseDataException, DataMalformedException


class Event(object):

    def __init__(self, id, name, slug, state, start_at, num_entrants,
                 check_in_buffer, check_in_duration, check_in_enabled,
                 is_online, team_name_allowed, team_management_deadline):
        self.id = id
        self.name = name
        self.slug = slug
        self.state = state
        self.start_at = start_at
        self.num_entrants = num_entrants
        self.check_in_buffer = check_in_buffer
        self.check_in_duration = check_in_duration
        self.check_in_enabled = check_in_enabled
        self.is_online = is_online
        self.team_name_allowed = team_name_allowed
        self.team_management_deadline = team_management_deadline
        self.identifier = '{}:{}'.format(self.id, self.name)

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.name, self.slug, self.state, self.start_at,
                     self.num_entrants, self.check_in_buffer, self.check_in_duration,
                     self.check_in_enabled, self.is_online, self.team_name_allowed,
                     self.team_name_allowed))

    @staticmethod
    def validate_data(input: dict, id: int=0):
        if 'data' in input:
            raise DataMalformedException(input)
        if 'event' not in input or input['event'] is None:
            raise NoEventDataException(id)

    @staticmethod
    def get(tournament_slug: str, event_slug: str):
        assert (tournament_slug is not None), 'Event.get cannot have None for tournament_slug parameter'
        assert (event_slug is not None), 'Event.get cannot have None for event_slug parameter'
        slug = "tournament/{0}/event/{1}".format(tournament_slug, event_slug)
        data = NI.query(queries.get_event_by_slugs, {"slug": slug})
        validate_data(data)

        try:
            event_data = data['data']['event']
            if event_data is None:
                raise NoEventDataException('{}:{}'.format(tournament_slug, event_slug))

            return Event.parse(event_data)
        except AttributeError as e:
            raise NoEventDataException('{}:{}'.format(tournament_slug, event_slug))

    @staticmethod
    def get_by_id(id: int):
        assert (id is not None), "Event.get_by_id cannot have None for id parameter"
        data = NI.query(queries.get_event_by_id, {'id': id})
        validate_data(data)

        try:
            event_data = data['data']['event']
            if event_data is None:
                raise NoEventDataException(id)

            return Event.parse(event_data)
        except AttributeError as e:
            raise NoEventDataException(id)

    @staticmethod
    def parse(data):
        assert (data is not None), 'Event.parse cannot have None for data parameter'
        if 'data' in data and 'event' in data['data']:
            raise DataMalformedException(data,
                                         'data is malformed for Event.parse. '
                                         'Please give only what is contained in the '
                                         '"event" property')

        assert ('id' in data), 'Event.parse must have id property on data parameter'
        assert ('name' in data), 'Event.parse must have name property on data parameter'
        assert ('slug' in data), 'Event.parse must have slug property on data parameter'
        assert ('state' in data), 'Event.parse must have state property on data parameter'
        assert ('startAt' in data), 'Event.parse must have startAt property on data parameter'
        assert ('numEntrants' in data), 'Event.parse must have numEntrants property on data parameter'
        assert ('checkInBuffer' in data), 'Event.parse must have checkInBuffer property on data parameter'
        assert ('checkInDuration' in data), 'Event.parse must have checkInDuration property on data parameter'
        assert ('checkInEnabled' in data), 'Event.parse must have checkInEnabled property on data parameter'
        assert ('isOnline' in data), 'Event.parse must have isOnline property on data parameter'
        assert ('teamNameAllowed' in data), 'Event.parse must have teamNameAllowed property on data parameter'
        assert ('teamManagementDeadline' in data), \
            'Event.parse must have teamManagementDeadline property on data parameter'
        return Event(
            data['id'],
            data['name'],
            data['slug'],
            data['state'],
            data['startAt'],
            data['numEntrants'],
            data['checkInBuffer'],
            data['checkInDuration'],
            data['checkInEnabled'],
            data['isOnline'],
            data['teamNameAllowed'],
            data['teamManagementDeadline']
        )

    def get_phases(self):
        assert (self.id is not None), 'event id cannot be None if calling get_phases'
        Logger.info('Getting Phases for Event: {0}:{1}'.format(self.id, self.name))
        data = NI.query(queries.get_event_phases, {'id': self.id})
        validate_data(data)

        try:
            event_data = data['data']['event']
            if event_data is None:
                raise NoEventDataException(self.identifier)

            phases_data = event_data['phases']
            if phases_data is None:
                raise NoPhaseDataException(self.identifier)

            return [Phase.parse(phase_data) for phase_data in phases_data]
        except AttributeError as e:
            raise Exception("No phase data pulled back for event {} {}".format(self.id, self.name))

    def get_phase_groups(self):
        assert (self.id is not None), 'event id cannot be None if calling get_phase_groups'
        Logger.info('Getting Phase Groups for Event: {0}:{1}'.format(self.id, self.name))
        data = NI.query(queries.get_event_phase_groups, {'id': self.id})
        validate_data(data)

        try:
            event_data = data['data']['event']
            if event_data is None:
                raise NoEventDataException(self.identifier)

            phase_groups_data = event_data['phaseGroups']
            if phase_groups_data is None:
                raise NoPhaseGroupDataException(self.identifier)

            return [PhaseGroup.parse(phase_group_data) for phase_group_data in phase_groups_data]
        except AttributeError as e:
            raise Exception("No phase group data pulled back for event {}".format(self.identifier))

    def get_attendees(self):
        assert (self.id is not None), 'event id cannot be None if calling get_attendees'
        Logger.info('Getting Attendees for Event: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        attendees = flatten([phase_group.get_attendees() for phase_group in phase_groups])
        return attendees

    def get_entrants(self):
        assert (self.id is not None), 'event id cannot be None if calling get_entrants'
        Logger.info('Getting Entrants for Event: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        entrants = flatten([phase_group.get_entrants() for phase_group in phase_groups])
        return entrants

    def get_sets(self):
        assert (self.id is not None), 'event id cannot be None if calling get_sets'
        Logger.info('Getting Sets for Event: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        sets = flatten([phase_group.get_sets() for phase_group in phase_groups])
        return sets

    def get_incomplete_sets(self):
        assert (self.id is not None), 'event id cannot be None if calling get_incomplete_sets'
        Logger.info('Getting Incomplete Sets for Event: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        incomplete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is False:
                incomplete_sets.append(ggset)
        return incomplete_sets

    def get_completed_sets(self):
        assert (self.id is not None), 'event id cannot be None if calling get_completed_sets'
        Logger.info('Getting Completed Sets for Event: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        complete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is True:
                complete_sets.append(ggset)
        return complete_sets

    # GETTERS
    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_slug(self):
        return self.slug

    def get_state(self):
        return self.state

    def get_start_at(self):
        return self.start_at

    def get_num_entrants(self):
        return self.num_entrants

    def get_check_in_buffer(self):
        return self.check_in_buffer

    def get_check_in_duration(self):
        return self.check_in_duration

    def get_check_in_enabled(self):
        return self.check_in_enabled

    def get_is_online(self):
        return self.is_online

    def get_team_name_allowed(self):
        return self.team_name_allowed

    def get_team_management_deadline(self):
        return self.team_management_deadline


from smashggpy.models.Phase import Phase
from smashggpy.models.PhaseGroup import PhaseGroup


from smashggpy.common.Exceptions import DataMalformedException, NoAttendeeDataException


class Attendee(object):

    def __init__(self, id, gamer_tag, prefix, created_at, claimed,
                 verified, player_id, phone_number, connected_accounts,
                 contact_info, event_ids):
        self.id = id
        self.gamer_tag = gamer_tag
        self.prefix = prefix
        self.created_at = created_at
        self.claimed = claimed
        self.verified = verified
        self.player_id = player_id
        self.phone_number = phone_number
        self.connected_accounts = connected_accounts
        self.contact_info = contact_info
        self.event_ids = event_ids

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.gamer_tag, self.prefix, self.created_at, self.claimed, self.verified,
                     self.player_id, self.phone_number, self.connected_accounts, self.contact_info, self.event_ids))

    @staticmethod
    def validate_data(input: dict, id: int=0) -> None:
        if 'data' in input:
            raise DataMalformedException(input)
        if 'phaseGroup' in input:
            input = input['phaseGroup']

        if 'participant' not in input and 'paginatedSeeds' not in input:
            raise NoAttendeeDataException(id)

        if 'participant' in input and input['participant'] is None:
            raise NoAttendeeDataException(id)
        elif 'paginatedSeeds' in input and input['paginatedSeeds']['nodes'] is None:
            raise NoAttendeeDataException(id)

    @staticmethod
    def parse(data):
        assert (data is not None), 'Attendee.parse must not have a none data parameter'
        assert ('id' in data), 'Attendee.parse must have a id property in data parameter'
        assert ('gamerTag' in data), 'Attendee.parse must have a gamerTag property in data parameter'
        assert ('prefix' in data), 'Attendee.parse must have a prefix property in data parameter'
        assert ('createdAt' in data), 'Attendee.parse must have a createdAt property in data parameter'
        assert ('claimed' in data), 'Attendee.parse must have a claimed property in data parameter'
        assert ('verified' in data), 'Attendee.parse must have a verified property in data parameter'
        assert ('playerId' in data), 'Attendee.parse must have a playerId property in data parameter'
        assert ('phoneNumber' in data), 'Attendee.parse must have a phoneNumber property in data parameter'
        assert ('connectedAccounts' in data), 'Attendee.parse must have a connectedAccounts property in data parameter'
        assert ('contactInfo' in data), 'Attendee.parse must have a contactInfo property in data parameter'
        assert ('events' in data), 'Attendee.parse must have a events property in data parameter'

        return Attendee(
            data['id'],
            data['gamerTag'],
            data['prefix'],
            data['createdAt'],
            data['claimed'],
            data['verified'],
            data['playerId'],
            data['phoneNumber'],
            data['connectedAccounts'],
            data['contactInfo'],
            [ids for ids in data['events']]
        )

    # GETTERS
    def get_id(self):
        return self.id
    
    def get_gamer_tag(self):
        return self.gamer_tag
    
    def get_prefix(self):
        return self.prefix
    
    def get_created_at(self):
        return self.created_at
    
    def get_claimed(self):
        return self.claimed
    
    def get_verified(self):
        return self.verified
    
    def get_player_id(self):
        return self.player_id
    
    def get_phone_number(self):
        return self.phone_number
    
    def get_connected_accounts(self):
        return self.connected_accounts
    
    def get_contact_info(self):
        return self.contact_info
    
    def get_event_ids(self):
        return self.event_ids
    



class QueryQueue(object):

	__instance = None
	__initialized = False

	def __init__(self, queue: list=[]):
		self.queue = queue
		

	@staticmethod
	def get_instance():
		if not QueryQueue.__initialized:
			QueryQueue.__instance = QueryQueue()
			QueryQueue.__initialized = True
		return QueryQueue.__instance

	@staticmethod
	def verify_initialized():
		if not QueryQueue.__initialized:
			raise NotInitializedException()

	def initialize(count: int):
		pass
		#Logger.debug('QueryQueue.initialize called')
		#for i in range(0, count, 1):
			#self.queue[i] = new QueryQueueElement()

	def get(self, index):
		Logger.debug('getting element from queue: {}'.format(index))
		return self.queue[index] or None

	def add(self, query):
		Logger.debug('adding element to queue: {}'.format(query))
		QueryQueue.verify_initialized()
		# thread = ThreadFactory.create(NI.execute_query, query)
		element = QueryQueueElement()
		self.queue.append(element)
		return element

	def pop(self):
		Logger.debug('popping from queue')
		QueryQueue.verify_initialized()
		return self.queue.pop(0)
		
	def length(self):
		# Logger.debug('getting length of Query Queue')
		return len(self.queue)

from smashggpy.util.Logger import Logger
from smashggpy.util.ThreadFactory import ThreadFactory
from smashggpy.util.QueryQueueElement import QueryQueueElement
from smashggpy.common.Exceptions import NotInitializedException
from smashggpy.util.NetworkInterface import NetworkInterface as NI



import smashggpy.queries.Schema as schema

phase_group_by_id = """
query PhaseGroupQueries($id: ID!){{
    phaseGroup(id: $id){{
        {0}
    }}
}}
""".format(schema.phase_group_schema)

phase_group_attendees = """
query PhaseGroupAttendees($id: ID!, $page: Int, $perPage: Int, $sortBy: String, $filter: SeedPageFilter){{
    phaseGroup(id: $id){{
        paginatedSeeds(
            query: {{
                page: $page,
                perPage: $perPage,
                sortBy: $sortBy,
                filter: $filter
            }}
        )
        {{
            pageInfo{{
                totalPages
            }}
            nodes{{
                entrant{{
                    participants{{
                        {0}
                    }}
                }}
            }}
        }}
    }}
}}
""".format(schema.attendee_schema)

phase_group_entrants = """
query PhaseGroupEntrants($id: ID!, $page: Int, $perPage: Int, $sortBy: String, $filter: SeedPageFilter){{
    phaseGroup(id: $id){{
        paginatedSeeds(
            query: {{
                page: $page,
                perPage: $perPage,
                sortBy: $sortBy,
                filter: $filter
            }}
        )
        {{
            pageInfo{{
                totalPages
            }}
            nodes{{
                entrant{{
                    {0}
                }}
            }}
        }}
    }}
}}
""".format(schema.entrant_schema)

phase_group_sets = """
query PhaseGroupEntrants($id: ID!, $page: Int, $perPage: Int, $sortType: SetSortType, $filters: SetFilters){{
    phaseGroup(id: $id){{
        paginatedSets(page:$page, perPage:$perPage, sortType:$sortType, filters:$filters){{
            pageInfo{{
                totalPages
            }}
            nodes{{
                {0}
            }}
        }}
    }}
}}
""".format(schema.gg_set_schema)



import smashggpy.queries.Phase_Queries as queries
from smashggpy.util.Logger import Logger
from smashggpy.common.Common import flatten, validate_data
from smashggpy.util.NetworkInterface import NetworkInterface as NI
from smashggpy.common.Exceptions import DataMalformedException, NoPhaseDataException, NoPhaseGroupDataException

class Phase(object):

    def __init__(self, id, name, num_seeds, group_count):
        self.id = id
        self.name = name
        self.num_seeds = num_seeds
        self.group_count = group_count

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.name, self.num_seeds, self.group_count))

    @staticmethod
    def validate_data(input, id: int=0):
        if 'data' in input:
            raise DataMalformedException(input)
        if 'phase' not in input or input['phase'] is None:
            raise NoPhaseDataException(id)

    @staticmethod
    def get(id: int):
        assert (id is not None), "Phase.get cannot have None for id parameter"
        data = NI.query(queries.phase_by_id, {'id': id})
        validate_data(data)
        Phase.validate_data(data['data'])

        base_data = data['data']['phase']
        return Phase.parse(base_data)

    @staticmethod
    def parse(data):
        assert (data is not None), 'Phase.parse cannot have None for data parameter'
        if 'data' in data and 'phase' in data['data']:
            raise DataMalformedException(data,
                                         'data is malformed for Phase.parse. '
                                         'Please give only what is contained in the '
                                         '"phase" property')

        assert ('id' in data), 'Phase.parse cannot have a None id property in data parameter'
        assert ('name' in data), 'Phase.parse cannot have a None name property in data parameter'
        assert ('numSeeds' in data), 'Phase.parse cannot have a None numSeeds property in data parameter'
        assert ('groupCount' in data), 'Phase.parse cannot have a None groupCount property in data parameter'

        return Phase(
            data['id'],
            data['name'],
            data['numSeeds'],
            data['groupCount']
        )

    def get_phase_groups(self):
        assert (self.id is not None), "phase id cannot be None when calling get_phase_groups"
        Logger.info('Getting Phase Groups for Phase: {0}:{1}'.format(self.id, self.name))
        data = NI.paginated_query(queries.phase_phase_groups, {'id': self.id})
        [validate_data(phase_data) for phase_data in data]

        # Schema Validation
        [Phase.validate_data(element['data'], self.id) for element in data]
        phase_data = [phase_data['data']['phase'] for phase_data in data]

        [PhaseGroup.validate_data(element, self.id) for element in phase_data]
        phase_group_data = flatten([element['phaseGroups'] for element in phase_data])

        return [PhaseGroup.parse(phase_group) for phase_group in phase_group_data]

    def get_attendees(self):
        Logger.info('Getting Attendees for Phase: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        attendees = flatten([phase_group.get_attendees() for phase_group in phase_groups])
        return attendees

    def get_entrants(self):
        Logger.info('Getting Entrants for Phase: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        entrants = flatten([phase_group.get_entrants() for phase_group in phase_groups])
        return entrants

    def get_sets(self):
        Logger.info('Getting Sets for Phase: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        sets = flatten([phase_group.get_sets() for phase_group in phase_groups])
        return sets

    def get_incomplete_sets(self):
        Logger.info('Getting Incomplete Sets for Phase: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        incomplete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is False:
                incomplete_sets.append(ggset)
        return incomplete_sets

    def get_completed_sets(self):
        Logger.info('Getting Completed Sets for Phase: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        complete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is True:
                complete_sets.append(ggset)
        return complete_sets

    # GETTERS
    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_num_seeds(self):
        return self.num_seeds

    def get_group_count(self):
        return self.group_count


from smashggpy.models.PhaseGroup import PhaseGroup



class Venue(object):

    def __init__(self, name, address, city, state, postal_code, country_code, region, latitude, longitude):
        self.name = name
        self.address = address
        self.city = city
        self.state = state
        self.postal_code = postal_code
        self.country_code = country_code
        self.region = region
        self.latitude = latitude
        self.longitude = longitude

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((self.name, self.address, self.city, self.state, self.postal_code,
                    self.country_code, self.region, self.latitude, self.longitude))

    @staticmethod
    def parse(data):
        assert (data is not None), 'Venue.parse must not have a None data parameter'
        assert ('venueName' in data), 'Venue.parse must have venueName property in data parameter'
        assert ('venueAddress' in data), 'Venue.parse must have a venueAddress property in data parameter'
        assert ('city' in data), 'Venue.parse must have a city property in data parameter'
        assert ('addrState' in data), 'Venue.parse must have an addrState property in data parameter'
        assert ('postalCode' in data), 'Venue.parse must have a postalCode property in data parameter'
        assert ('countryCode' in data), 'Venue.parse must have a countryCode property in data parameter'
        assert ('region' in data), 'Venue.parse must have a region property in data parameter'
        assert ('lat' in data), 'Venue.parse must have a lat property in data parameter'
        assert ('lng' in data), 'Venue.parse must have a lng property in data parameter'
        return Venue(
            data['venueName'],
            data['venueAddress'],
            data['city'],
            data['addrState'],
            data['postalCode'],
            data['countryCode'],
            data['region'],
            data['lat'],
            data['lng']
        )

    # GETTERS
    def get_name(self):
        return self.name

    def get_address(self):
        return self.address

    def get_city(self):
        return self.city

    def get_state(self):
        return self.state

    def get_postal_code(self):
        return self.postal_code

    def get_country_code(self):
        return self.country_code

    def get_region(self):
        return self.region

    def get_latitude(self):
        return self.latitude

    def get_longitude(self):
        return self.longitude



import time
from threading import Thread

class QueryQueueElement(object):

	def __init__(self, timestamp=None):
		self.timestamp = timestamp

	def set_timestamp(self):
		self.timestamp = time.time()

	def get_time_difference_in_seconds(self):
		now = time.time()
		return now - self.timestamp

from smashggpy.util.Logger import Logger


from threading import Thread

class ThreadFactory(object):

	@staticmethod
	def create(function, argument_dict):
		return Thread(group=None, target=function, name=None, args=(), kwargs=argument_dict)

from smashggpy.util.Logger import Logger


import time

class QueryFactory(object):

	@staticmethod
	def create(query: str, variables: dict):
		Logger.get_instance().debug('created query with params [{}] [{}]'.format(query, variables))
		return Query(query, variables)


from smashggpy.util.Query import Query
from smashggpy.util.Logger import Logger



class StreamQueue(object):

    def __init__(self, stream, sets):
        self.stream = stream
        self.sets = sets

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.stream, self.sets))

    @staticmethod
    def parse(data):
        return StreamQueue(

        )

    def get_stream(self):
        return self.stream

    def get_sets(self):
        return self.sets



class Entrant(object):

    def __init__(self, id, name, event_id, skill, attendee_data):
        self.id = id
        self.name = name
        self.event_id = event_id
        self.skill = skill
        self.attendee_data = attendee_data

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.name, self.event_id, self.skill, self.attendee_data))

    @staticmethod
    def parse(data):
        assert (data is not None), 'Entrant.parse must not have a none data parameter'
        assert ('id' in data), 'Entrant.parse must have a id property in data parameter'
        assert ('name' in data), 'Entrant.parse must have a name property in data parameter'
        assert ('eventId' in data), 'Entrant.parse must have a eventId property in data parameter'
        assert ('skill' in data), 'Entrant.parse must have a skill property in data parameter'
        assert ('participants' in data), 'Entrant.parse must have a participants property in data parameter'

        return Entrant(
            data['id'],
            data['name'],
            data['eventId'],
            data['skill'],
            [Attendee.parse(attendee_data) for attendee_data in data['participants']]
        )

    # GETTERS
    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_event_id(self):
        return self.event_id

    def get_skill(self):
        return self.skill

    def get_attendee_data(self):
        return self.attendee_data


from smashggpy.models.Attendee import Attendee


import smashggpy.queries.Tournament_Queries as queries

from smashggpy.util.Logger import Logger
from smashggpy.common.Common import flatten, validate_data
from smashggpy.util.NetworkInterface import NetworkInterface as NI
from smashggpy.models.Venue import Venue
from smashggpy.models.Organizer import Organizer
from smashggpy.common.Exceptions import \
    DataPullException, NoTournamentDataException, NoEventDataException, \
    NoPhaseDataException, NoPhaseGroupDataException, DataMalformedException


class Tournament(object):

    def __init__(self, id, name, slug, start_time, end_time, timezone, venue, organizer):
        self.id = id
        self.name = name
        self.slug = slug
        self.start_time = start_time
        self.end_time = end_time
        self.timezone = timezone
        self.venue = venue
        self.organizer = organizer

    def __eq__(self, other):
        if other is None:
            return False

        if type(self) != type(other):
            return False

        return hash(self) == hash(other)

    def __hash__(self):
        return hash((self.id, self.name, self.slug, self.start_time,
                     self.end_time, self.timezone, self.venue, self.organizer))

    @staticmethod
    def validate_data(input: dict, slug: str = None):
        if 'data' in input:
            raise DataMalformedException(input)
        if 'tournament' not in input or input['tournament'] is None:
            raise NoTournamentDataException(slug)

    @staticmethod
    def get(slug: str):
        assert (slug is not None), "Tournament.get must have a slug parameter"
        data = NI.query(queries.get_tournament_by_slug, {'slug': slug})
        validate_data(data)

        try:
            tournament_data = data['data']['tournament']
            if tournament_data is None:
                raise NoTournamentDataException(slug)

            return Tournament.parse(tournament_data)
        except AttributeError as e:
            raise NoTournamentDataException(slug)

    @staticmethod
    def get_by_id(id: int):
        assert (id is not None), "Tournament.get_by_id must have an id parameter"
        data = NI.query(queries.get_tournament, {'id': id})
        validate_data(data)

        try:
            tournament_data = data['data']['tournament']
            if tournament_data is None:
                raise NoTournamentDataException(id)

            return Tournament.parse(tournament_data)
        except AttributeError as e:
            raise NoTournamentDataException(id)

    @staticmethod
    def parse(data):
        assert (data is not None), 'Tournament.parse must have a data parameter'
        if 'data' in data and 'tournament' in data['data']:
            raise DataMalformedException(data,
                                         'data is malformed for Tournament.parse. '
                                         'Please give only what is contained in the '
                                         '"tournament" property')

        assert ('id' in data), 'Tournament.parse must have id in data parameter'
        assert ('name' in data), 'Tournament.parse must have id in data parameter'
        assert ('slug' in data), 'Tournament.parse must have id in data parameter'
        assert ('startAt' in data), 'Tournament.parse must have id in data parameter'
        assert ('endAt' in data), 'Tournament.parse must have id in data parameter'
        assert ('timezone' in data), 'Tournament.parse must have id in data parameter'

        return Tournament(
            data['id'],
            data['name'],
            data['slug'],
            data['startAt'],
            data['endAt'],
            data['timezone'],
            Venue.parse(data),
            Organizer.parse(data)
        )

    def get_events(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_events"
        data = NI.query(queries.get_tournament_events, {'id': self.id})
        validate_data(data)

        tournament_data = data['data']['tournament']
        if tournament_data is None:
            raise NoTournamentDataException(self.slug)

        base_data = tournament_data['events']
        if base_data is None:
            raise NoEventDataException(self.slug)

        return [Event.parse(event_data) for event_data in base_data]

    def get_phases(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_phases"

        phases = []
        data = NI.query(queries.get_tournament_phases, {'id': self.id})
        validate_data(data)

        tournament_data = data['data']['tournament']
        if tournament_data is None:
            raise NoTournamentDataException(self.slug)

        event_data = tournament_data['events']
        if event_data is None:
            raise NoEventDataException(self.slug)

        phase_data = [event['phases'] for event in event_data]
        for phase in phase_data:
            if phase is None:
                raise NoPhaseDataException(self.slug)
            [phases.append(Phase.parse(event_phase)) for event_phase in phase]

        return phases

    def get_phase_groups(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_phase_groups"

        phase_groups = []
        data = NI.query(queries.get_tournament_phase_groups, {'id': self.id})
        validate_data(data)

        tournament_data = data['data']['tournament']
        if tournament_data is None:
            raise NoTournamentDataException(self.slug)

        event_data = tournament_data['events']
        if event_data is None:
            raise NoEventDataException(self.slug)

        phase_groups_data = [event['phaseGroups'] for event in event_data]
        for phase_group in phase_groups_data:
            if phase_group is None:
                raise NoPhaseGroupDataException(self.slug)
            [phase_groups.append(PhaseGroup.parse(phase_group_data)) for phase_group_data in phase_group]

        return phase_groups

    def get_attendees(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_attendees"
        Logger.info('Getting Attendees for Tournament: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        attendees = flatten([phase_group.get_attendees() for phase_group in phase_groups])
        return attendees

    def get_entrants(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_entrants"
        Logger.info('Getting Entrants for Tournament: {0}:{1}'.format(self.id, self.name))
        Logger.warning('Aggregate queries ')
        phase_groups = self.get_phase_groups()
        entrants = flatten([phase_group.get_entrants() for phase_group in phase_groups])
        return entrants

    def get_sets(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_sets"
        Logger.info('Getting Sets for Tournament: {0}:{1}'.format(self.id, self.name))
        phase_groups = self.get_phase_groups()
        sets = flatten([phase_group.get_sets() for phase_group in phase_groups])
        return sets

    def get_incomplete_sets(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_incomplete_sets"
        Logger.info('Getting Incomplete Sets for Tournament: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        incomplete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is False:
                incomplete_sets.append(ggset)
        return incomplete_sets

    def get_completed_sets(self):
        assert (self.id is not None), "tournament id cannot be None if calling get_completed_sets"
        Logger.info('Getting Completed Sets for Tournament: {0}:{1}'.format(self.id, self.name))
        sets = self.get_sets()
        complete_sets = []
        for ggset in sets:
            if ggset.get_is_completed() is True:
                complete_sets.append(ggset)
        return complete_sets

    # GETTERS
    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_slug(self):
        return self.slug

    def get_start_time(self):
        return self.start_time

    def get_end_time(self):
        return self.end_time

    def get_timezone(self):
        return self.timezone

    def get_venue(self):
        return self.venue

    def get_organizer(self):
        return self.organizer


from smashggpy.models.Event import Event
from smashggpy.models.Phase import Phase
from smashggpy.models.PhaseGroup import PhaseGroup



import re

from smashggpy.util.Logger import Logger

class TokenHandler(object):

	__token = None
	__initialized = False
	__regex = re.compile('[a-f0-9]{32}')

	@staticmethod
	def init(token: str):
		log = Logger.get_instance()
		
		if not TokenHandler.__initialized:
			log.info('Checking API token: {}...'.format(token))
			if not TokenHandler.__regex.match(token):
				raise BadTokenException(token)
			elif len(token) > 32:
				raise BadTokenException(token)

			TokenHandler.__token = token
			TokenHandler.__initialized = True
			log.info('Legal API Token!')

	@staticmethod
	def get_token():
		if not TokenHandler.__initialized:
			raise NotInitializedException()
		return TokenHandler.__token

	@staticmethod
	def uninit():
		TokenHandler.__token = None
		TokenHandler.__initialized = False

from smashggpy.common.Exceptions import NotInitializedException, BadTokenException


import json
import time

class Query(object):

	def __init__(self, query: str='', variables: dict={}):
		self.query = query
		self.variables = variables

	def get_query_dict(self):
		return {
			'query': self.query,
			'variables': json.dumps(self.variables)
		}

	def get_query_string(self):
		return json.dumps({
			'query': self.query,
			'variables': json.dumps(self.variables)
		})

	def get_query(self):
		return self.query

	def get_query_s(self):
		return str(self.query)

	def get_variables(self):
		return self.variables

	def get_variables_s(self):
		return str(self.variables)

from smashggpy.util.Logger import Logger


class NotInitializedException(Exception):
    def __init__(self, component_name):
        self.message = '{} has not been initialized!'.format(component_name)


class BadTokenException(Exception):
    def __init__(self, token='No token handed to Exception'):
        self.message = 'Bad Token. Please input a 32 character hexidecimal string.\nGot input "{}"'.format(token)


class NoTournamentDataException(Exception):
    def __init__(self, slug: str):
        self.message = 'No tournament data pulled back for tournament {}'.format(slug)


class NoEventDataException(Exception):
    def __init__(self, slug: str):
        self.message = 'No event data pulled back for tournament {}'.format(slug)


class NoPhaseDataException(Exception):
    def __init__(self, slug: str):
        self.message = 'No phase data pulled back for tournament {}'.format(slug)


class NoPhaseGroupDataException(Exception):
    def __init__(self, slug: str):
        self.message = 'No phase group data pulled back for tournament {}'.format(slug)


class NoAttendeeDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class NoEntrantDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class NoUserDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class NoSetDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class NoPlayerDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class NoStreamDataException(Exception):
    def __init__(self, id: int):
        self.message = 'No phase group data pulled back for tournament {}'.format(id)


class DataPullException(Exception):
    def __init__(self, identifier: str, msg: str):
        self.message = 'An error occurred when pulling data for {}: {}'.format(identifier, msg)


class DataMalformedException(Exception):
    def __init__(self, data: dict, msg: str = None):
        if msg is None:
            self.message = 'Data came back malformed: {}'.format(data)
        else:
            self.message = '{}: {}'.format(msg, data)



import json
import requests
from smashggpy.common.Common import flatten

class NetworkInterface(object):

	API_URL='https://api.smash.gg/gql/alpha'

	@staticmethod
	def get_headers():
		return {
			'X-Source': 'smashggpy',
			'Content-Type': 'application/json',
			'Authorization': 'Bearer {}'.format(TokenHandler.get_token())
		}

	@staticmethod
	def query(query_string: str, variables: dict):
		Logger.debug('NetworkInterface.query: creating query object')
		query = QueryFactory.create(query_string, variables)
		Logger.debug('NetworkInterface.query: created query {}'.format(query))

		Logger.debug('NetworkInterface.query: sending query to queue')
		QueryQueue.get_instance().add(query)
		return NetworkInterface.execute_query(query)

	@staticmethod
	def paginated_query(query_string: str, variables: dict) -> list:
		Logger.debug('NetworkInterface.paginated_query: creating query object')
		query = QueryFactory.create(query_string, variables)
		Logger.debug('NetworkInterface.paginated_query: created query {}'.format(query))

		Logger.debug('NetworkInterface.paginated_query: sending query to queue')
		QueryQueue.get_instance().add(query)

		results = []
		initial_result = NetworkInterface.execute_query(query)
		base_data = NetworkInterface.parse_paginated_result(initial_result)
		results.append(base_data['nodes'])

		total_pages = base_data['pageInfo']['totalPages']
		for i in range(2, total_pages + 1, 1):
			variables['page'] = i
			current_query = QueryFactory.create(query_string, variables)
			current_result = NetworkInterface.execute_query(current_query)
			current_base_data = NetworkInterface.parse_paginated_result(current_result)
			results.append(current_base_data['nodes'])

		return flatten(results)

	@staticmethod
	def parse_paginated_result(results: dict):
		main_key = list(results['data'].keys())[0]
		secondary_key = list(results['data'][main_key].keys())[0]
		base_data = results['data'][main_key][secondary_key]
		return base_data

	@staticmethod
	def execute_query(query):
		log = Logger.get_instance()
		url = NetworkInterface.API_URL
		headers = NetworkInterface.get_headers()
		payload = query.get_query_dict()

		log.debug('NetworkInterface.query: Payload: {}'.format(payload))
		log.debug('NetworkInterface.query: Headers: {}'.format(headers))

		response = requests.post(url=url, headers=headers, json=payload)

		log.debug('NetworkInterface.query: {}'.format(response))
		log.debug('NetworkInterface.query: JSON Response: {}'.format(response.json()))
		return response.json()

# Path imports
from smashggpy.util.Logger import Logger
from smashggpy.util.TokenHandler import TokenHandler
from smashggpy.util.QueryFactory import QueryFactory
from smashggpy.util.QueryQueue import QueryQueue


import smashggpy.queries.Phase_Group_Queries as queries
from smashggpy.common.Common import flatten, validate_data
from smashggpy.util.Logger import Logger
from smashggpy.util.NetworkInterface import NetworkInterface as NI
from smashggpy.common.Exceptions import DataMalformedException, NoPhaseGroupDataException

class PhaseGroup(object):

    def __init__(self, id, display_identifier, first_round_time, state, phase_id, wave_id, tiebreak_order):
        self.id = id
        self.display_identifier = display_identifier
        self.first_round_time = first_round_time
        self.state = state
        self.phase_id = phase_id
        self.wave_id = wave_id
        self.tiebreak_order = tiebreak_order

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.display_identifier, self.first_round_time,
                     self.state, self.phase_id, self.wave_id,
                     frozenset(self.tiebreak_order) if self.tiebreak_order is not None else None))

    @staticmethod
    def validate_data(input, id: int=0):
        if 'data' in input:
            raise DataMalformedException(input,
                                         'data is malformed for PhaseGroup.parse. '
                                         'Please give only what is contained in the '
                                         '"phaseGroup" property')
        if 'phaseGroup' not in input and 'phaseGroups' not in input:
            raise NoPhaseGroupDataException(id)
        elif 'phaseGroup' in input and input['phaseGroup'] is None:
            raise NoPhaseGroupDataException(id)
        elif 'phaseGroups' in input and input['phaseGroups'] is None:
            raise NoPhaseGroupDataException(id)

    @staticmethod
    def get(id: int):
        assert (id is not None), "PhaseGroup.get cannot have None for id parameter"
        data = NI.query(queries.phase_group_by_id, {'id': id})
        validate_data(data)
        PhaseGroup.validate_data(data['data'])

        base_data = data['data']['phaseGroup']
        return PhaseGroup.parse(base_data)

    @staticmethod
    def parse(data):
        assert (data is not None), "PhaseGroup.parse cannot have None for data parameter"
        if 'data' in data:
            raise DataMalformedException(data,
                                         'data is malformed for PhaseGroup.parse. '
                                         'Please give only what is contained in the '
                                         '"phaseGroup" property')

        assert ('id' in data), 'PhaseGroup.parse must have a id property in data parameter'
        assert ('displayIdentifier' in data), \
            'PhaseGroup.parse cannot must have a displayIdentifier property in data parameter'
        assert ('firstRoundTime' in data), \
            'PhaseGroup.parse cannot must have a firstRoundTime property in data parameter'
        assert ('state' in data), 'PhaseGroup.parse cannot must have a state property in data parameter'
        assert ('phaseId' in data), 'PhaseGroup.parse cannot must have a phaseId property in data parameter'
        assert ('waveId' in data), 'PhaseGroup.parse cannot must have a waveId property in data parameter'
        assert ('tiebreakOrder' in data), 'PhaseGroup.parse cannot must have a tiebreakOrder property in data parameter'

        return PhaseGroup(
            data['id'],
            data['displayIdentifier'],
            data['firstRoundTime'],
            data['state'],
            data['phaseId'],
            data['waveId'],
            data['tiebreakOrder']
        )

    def get_attendees(self):
        assert (self.id is not None), 'phase group id cannot be None when calling get_attendees'
        Logger.info('Getting Attendees for phase group: {0}:{1}'.format(self.id, self.display_identifier))
        data = NI.paginated_query(queries.phase_group_attendees, {'id': self.id})
        validate_data(data)

        participants = flatten([entrant_data['entrant']['participants'] for entrant_data in data])
        attendees = [Attendee.parse(participant_data) for participant_data in participants]
        return attendees

    def get_entrants(self):
        assert (self.id is not None), 'phase group id cannot be None when calling get_entrants'
        Logger.info('Getting Entrants for phase group: {0}:{1}'.format(self.id, self.display_identifier))
        data = NI.paginated_query(queries.phase_group_entrants, {'id': self.id})
        validate_data(data)
        entrants = [Entrant.parse(entrant_data['entrant']) for entrant_data in data]
        return entrants

    def get_sets(self):
        assert (self.id is not None), 'phase group id cannot be None when calling get_sets'
        Logger.info('Getting Sets for phase group: {0}:{1}'.format(self.id, self.display_identifier))
        data = NI.paginated_query(queries.phase_group_sets, {'id': self.id})
        validate_data(data)
        sets = [GGSet.parse(set_data) for set_data in data]
        return sets

    def get_incomplete_sets(self):
        assert (self.id is not None), 'phase group id cannot be None when calling get_incomplete_sets'
        Logger.info('Getting Incomplete Sets for phase group: {0}:{1}'.format(self.id, self.display_identifier))
        sets = self.get_sets()
        incomplete_sets = []
        for gg_set in sets:
            if set.get_is_completed() is False:
                incomplete_sets.append(gg_set)
        return incomplete_sets

    def get_completed_sets(self):
        assert (self.id is not None), 'phase group id cannot be None when calling get_completed_sets'
        Logger.info('Getting Completed Sets for phase group: {0}:{1}'.format(self.id, self.display_identifier))
        sets = self.get_sets()
        incomplete_sets = []
        for gg_set in sets:
            if set.get_is_completed() is True:
                incomplete_sets.append(gg_set)
        return incomplete_sets

    # GETTERS
    def get_id(self):
        return self.id

    def get_display_identifier(self):
        return self.display_identifier

    def get_first_round_time(self):
        return self.first_round_time

    def get_state(self):
        return self.state

    def get_phase_id(self):
        return self.phase_id

    def get_wave_id(self):
        return self.wave_id

    def get_tiebreak_order(self):
        return self.tiebreak_order


from smashggpy.models.Entrant import Entrant
from smashggpy.models.Attendee import Attendee
from smashggpy.models.GGSet import GGSet




__daemon_thread = None

def initialize(api_token: str, log_level: str='info'):
	Logger.init(log_level)
	TokenHandler.init(api_token)

	# initialize the query queue
	# QueryQueue.init()

	# initialize query queue daemon in background
	QueryQueueDaemon.run_daemon()

def uninitialize():
	TokenHandler.uninit()
	QueryQueueDaemon.kill_daemon()

from smashggpy.util.Logger import Logger
from smashggpy.util.TokenHandler import TokenHandler
from smashggpy.util.QueryQueue import QueryQueue
from smashggpy.util.ThreadFactory import ThreadFactory
from smashggpy.util.QueryQueueDaemon import QueryQueueDaemon



class Stream(object):

    def __init__(self, id, event_id, tournament_id, stream_name, num_setups,
                 stream_source, stream_type, stream_type_id, is_online,
                 enabled, follower_count, removes_tasks, stream_status,
                 stream_game, stream_logo):
        self.id = id
        self.event_id = event_id
        self.tournament_id = tournament_id
        self.stream_name = stream_name
        self.num_setups = num_setups
        self.stream_source = stream_source
        self.stream_type = stream_type
        self.stream_type_id = stream_type_id
        self.is_online = is_online
        self.enabled = enabled
        self.follower_count = follower_count
        self.removes_tasks = removes_tasks
        self.stream_status = stream_status
        self.stream_game = stream_game
        self.stream_logo = stream_logo

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.event_id, self.tournament_id, self.stream_name, self.num_setups,
                     self.stream_source, self.stream_type, self.stream_type_id, self.is_online,
                     self.enabled, self.follower_count, self.removes_tasks, self.stream_status,
                     self.stream_game, self.stream_logo))

    @staticmethod
    def parse(data):
        assert (data is not None), 'Stream.parse must not have a none data parameter'
        assert ('id' in data), 'Stream.parse must have a id property in data parameter'
        assert ('eventId' in data), 'Stream.parse must have a eventId property in data parameter'
        assert ('tournamentId' in data), 'Stream.parse must have a tournamentId property in data parameter'
        assert ('streamName' in data), 'Stream.parse must have a streamName property in data parameter'
        assert ('numSetups' in data), 'Stream.parse must have a numSetups property in data parameter'
        assert ('streamSource' in data), 'Stream.parse must have a streamSource property in data parameter'
        assert ('streamType' in data), 'Stream.parse must have a streamType property in data parameter'
        assert ('streamTypeId' in data), 'Stream.parse must have a streamTypeId property in data parameter'
        assert ('isOnline' in data), 'Stream.parse must have a isOnline property in data parameter'
        assert ('enabled' in data), 'Stream.parse must have a enabled property in data parameter'
        assert ('followerCount' in data), 'Stream.parse must have a followerCount property in data parameter'
        assert ('removesTasks' in data), 'Stream.parse must have a removesTasks property in data parameter'
        assert ('streamStatus' in data), 'Stream.parse must have a streamStatus property in data parameter'
        assert ('streamGame' in data), 'Stream.parse must have a streamGame property in data parameter'
        assert ('streamLogo' in data), 'Stream.parse must have a streamLogo property in data parameter'

        return Stream(
            data['id'],
            data['eventId'],
            data['tournamentId'],
            data['streamName'],
            data['numSetups'],
            data['streamSource'],
            data['streamType'],
            data['streamTypeId'],
            data['isOnline'],
            data['enabled'],
            data['followerCount'],
            data['removesTasks'],
            data['streamStatus'],
            data['streamGame'],
            data['streamLogo'],
        )

    # GETTERS
    def get_id(self):
        return self.id
    
    def get_event_id(self):
        return self.event_id
    
    def get_tournament_id(self):
        return self.tournament_id
    
    def get_stream_name(self):
        return self.stream_name
    
    def get_num_setups(self):
        return self.num_setups
    
    def get_stream_source(self):
        return self.stream_source
    
    def get_stream_type(self):
        return self.stream_type
    
    def get_stream_type_id(self):
        return self.stream_type_id
    
    def get_is_online(self):
        return self.is_online
    
    def get_enabled(self):
        return self.enabled
    
    def get_follower_count(self):
        return self.follower_count
    
    def get_removes_tasks(self):
        return self.removes_tasks
    
    def get_stream_status(self):
        return self.stream_status
    
    def get_stream_game(self):
        return self.stream_game
    
    def get_stream_logo(self):
        return self.stream_logo
    


tournament_schema = """
id
name
slug
city
postalCode
addrState
countryCode
region
venueAddress
venueName
gettingThere
lat
lng
timezone
startAt
endAt
contactEmail
contactTwitter
contactPhone
ownerId
"""

event_schema = """
id
name
slug
state
startAt
numEntrants
checkInBuffer
checkInDuration
checkInEnabled
isOnline
teamNameAllowed
teamManagementDeadline
"""

phase_schema = """
id
name
numSeeds
groupCount
"""

phase_group_schema = """
id
displayIdentifier
firstRoundTime
state
phaseId
waveId
tiebreakOrder
"""

user_schema = """
id
gamerTag
prefix
color
twitchStream
twitterHandle
youtube
region
state
country
gamerTagChangedAt
"""

attendee_contact_info_schema = """
id
city
state
stateId
country
countryId
name
nameFirst
nameLast
zipcode
"""

attendee_schema = """
id
gamerTag
prefix
createdAt
claimed
verified
playerId
phoneNumber
connectedAccounts
contactInfo{{
    {0}
}}
events{{
    id
}}
""".format(attendee_contact_info_schema)

entrant_schema = """
id
name
eventId
skill
participants{{
    {0}
}}
""".format(attendee_schema)

gg_set_slot_schema = """
id
entrant {
    id
    name
    participants {
        id
    }
}
"""

gg_set_schema = """
id
eventId
phaseGroupId
displayScore
fullRoundText
round
startedAt
completedAt
winnerId
totalGames
state
slots(includeByes: false){{
    {0}
}}
""".format(gg_set_slot_schema)


import os
from pathlib import Path  
from dotenv import load_dotenv

from smashggpy.common.Exceptions import DataPullException, DataMalformedException

BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
ROOT_DIR = BASE_DIR/'..'/'..'


def dotenv():
	env_path = ROOT_DIR / '.env'
	load_dotenv(dotenv_path=env_path)


def flatten(l: list):
	flat_list = []
	for sublist in l:
		for item in sublist:
			flat_list.append(item)
	return flat_list


def validate_data(data: dict):
	if 'data' not in data:
		raise DataMalformedException(data)

	if 'errors' in data['data']:
		raise DataPullException(id, data['data']['errors'])



import time
from smashggpy.util.ThreadFactory import ThreadFactory

class QueryQueueDaemon(object):
	'''
	This class is a loop that runs forever on a seperate thread. This 
	class continuously checks the QueryQueue Singleton to verify it
	is following the Rate Limiting rules of Smashgg.

	When the Rate Limit amount of elements in the Queue has been reached,
	the rest of the Queries that come in will be Queued but not executed.

	The Daemon will check when the elements at the front have surpassed the 
	amount of time necessary before a new Query may be executed and then 
	lets the newest query into Non-Delinquency to be executed. At this time,
	a timestamp is appended to it so as we may evaluate the length of time
	that has passed since the query was added to the Queue.
	'''

	__keep_alive = True
	__daemon_thread = None

	@staticmethod
	def kill_daemon():
		QueryQueueDaemon.__keep_alive = False
		QueryQueueDaemon.__daemon_thread.join()

	
	@staticmethod
	def run_daemon(DELINQUENCY_RATE: int=80, QUERY_TIME_IN_SECONDS: int=60):
		QueryQueueDaemon.__keep_alive = True
		QueryQueueDaemon.__daemon_thread = ThreadFactory.create(
			QueryQueueDaemon.daemon,
			{
				'DELINQUENCY_RATE': DELINQUENCY_RATE,
				'QUERY_TIME_IN_SECONDS': QUERY_TIME_IN_SECONDS
			}
		)
		QueryQueueDaemon.__daemon_thread.start()

	@staticmethod
	def daemon(DELINQUENCY_RATE: int=80, QUERY_TIME_IN_SECONDS: int=60):
		"""
		1) check if the queue is empty, if not check if element 0 should be popped
		2) check if we are delinquent and if so give the user a message about how long 
			until the next query is fired
		3) check if the queue has reached unexecuted non-delinquent queries
		"""
		active_threads = []

		log = Logger.get_instance()
		log.info('Running the QueryQueue Daemon')

		while QueryQueueDaemon.__keep_alive is True:
			log = Logger.get_instance()
			now = time.time()
			queue = QueryQueue.get_instance()

			if queue.length() > 0:
				queue_length = queue.length()
				back_limit = DELINQUENCY_RATE - 1 if queue_length > DELINQUENCY_RATE else queue_length - 1
				Logger.debug('Queue size: {}, Back Limit: {}'.format(queue_length, back_limit))

				# execute waiting non-delinquent queries and fill in missing timestamps
				for i in range(back_limit, -1, -1):
					current_element = queue.get(i)
					if current_element.timestamp is None:
						current_element.set_timestamp()

				# determine if we need to pop elements
				front_element = queue.get(0)
				if front_element is None: continue
				time_difference_in_seconds = now - front_element.timestamp
				Logger.debug('time difference in seconds: {}'.format(time_difference_in_seconds))
				if time_difference_in_seconds > QUERY_TIME_IN_SECONDS:
					Logger.debug('removing element')
					queue.pop()
						

from smashggpy.util.Logger import Logger
from smashggpy.util.QueryQueue import QueryQueue
from smashggpy.util.ThreadFactory import ThreadFactory
from smashggpy.util.NetworkInterface import NetworkInterface as NI
from smashggpy.util.QueryQueueElement import QueryQueueElement



class User(object):

    def __init__(self, id, gamer_tag, prefix, color, twitch_stream,
                 twitter_handle, youtube, region, state, country,
                 gamer_tag_changed_at):
        self.id = id
        self.gamer_tag = gamer_tag
        self.prefix = prefix
        self.color = color
        self.twitch_stream = twitch_stream
        self.twitter_handle = twitter_handle
        self.youtube = youtube
        self.region = region
        self.state = state
        self.country = country
        self.gamer_tag_changed_at = gamer_tag_changed_at

    def __eq__(self, other):
        if other is None:
            return False
        if type(other) != type(self):
            return False
        return hash(other) == hash(self)

    def __hash__(self):
        return hash((self.id, self.gamer_tag, self.prefix, self.color, self.twitch_stream,
                     self.twitter_handle, self.youtube, self.region, self.state, self.country,
                     self.gamer_tag_changed_at))

    @staticmethod
    def parse(data):
        assert (data is not None), 'User.parse must not have a none data parameter'
        assert ('id' in data), 'User.parse must have a id property in data parameter'
        assert ('gamerTag' in data), 'User.parse must have a gamerTag property in data parameter'
        assert ('prefix' in data), 'User.parse must have a prefix property in data parameter'
        assert ('color' in data), 'User.parse must have a color property in data parameter'
        assert ('twitchStream' in data), 'User.parse must have a twitchStream property in data parameter'
        assert ('twitterHandle' in data), 'User.parse must have a twitterHandle property in data parameter'
        assert ('youtube' in data), 'User.parse must have a youtube property in data parameter'
        assert ('region' in data), 'User.parse must have a region property in data parameter'
        assert ('state' in data), 'User.parse must have a state property in data parameter'
        assert ('country' in data), 'User.parse must have a country property in data parameter'
        assert ('gamerTagChangedAt' in data), 'User.parse must have a gamerTagChangedAt property in data parameter'

        return User(
            data['id'],
            data['gamerTag'],
            data['prefix'],
            data['color'],
            data['twitchStream'],
            data['twitterHandle'],
            data['youtube'],
            data['region'],
            data['state'],
            data['country'],
            data['gamerTagChangedAt']
        )

    # GETTERS
    def get_id(self):
        return self.id

    def get_gamer_tag(self):
        return self.gamer_tag

    def get_prefix(self):
        return self.prefix

    def get_color(self):
        return self.color

    def get_twitch_stream(self):
        return self.twitch_stream

    def get_twitter_handle(self):
        return self.twitter_handle

    def get_youtube(self):
        return self.youtube

    def get_region(self):
        return self.region

    def get_state(self):
        return self.state

    def get_country(self):
        return self.country

    def get_gamer_tag_changed_at(self):
        return self.gamer_tag_changed_at



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="smashggpy",
    version="0.0.27",
    author="Brandon Cooke",
    author_email="brandoncookedev@gmail.com",
    description="Python SDK for smash.gg's public api",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BrandonCookeDev/smashgg.py",
    packages=['smashggpy.util', 'smashggpy.models', 'smashggpy.queries', 'smashggpy.common'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="GNU Public License v3.0",
    keywords='smashgg smash.gg smashgg.py smashggpy sdk wrapper api gql graphql',
    python_requires='~=3.7',
)




import smashggpy.queries.Schema as schema

get_event_by_id = """
query EventQuery($id: ID!){{
    event(id: $id){{
        {0}
    }}
}}
""".format(schema.event_schema)

get_event_by_slugs = """
query EventQuery($slug: String){{
    event(slug: $slug){{
        {0}
    }}
}}
""".format(schema.event_schema)

get_event_phases = """
query EventPhases($id: ID!){{
    event(id: $id){{
        phases{{
            {0}
        }}
    }}
}}
""".format(schema.phase_schema)

get_event_phase_groups = """
query EventPhaseGroups($id: ID!){{
    event(id: $id){{
        phaseGroups{{
            {0}
        }}
    }}
}}
""".format(schema.phase_group_schema)